import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.geometry.BoundingBox;
import javafx.geometry.Point2D;
import javafx.scene.Group;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.image.Image;
import javafx.scene.paint.Color;
import javafx.scene.text.Font;
import javafx.scene.text.FontPosture;
import javafx.scene.transform.Affine;
import javafx.stage.Stage;
import javafx.util.Duration;

/**
 * BigForest.java
 *
 * Demo 2D scrolling
 *
 * @author mike slattery
 * @version feb 2016
 */
public class BigForest extends Application {
	final String appName = "BigForest";
	final int FPS = 30; // frames per second
	public static final int DYING = 80;

	TreeSprite tree;
	FireSprite fire;
	Sprite scenery[];
	Point2D dots[];
	boolean boxDebug = false;
	Image rockImage, lakeImage, treeImage, fireImage;
	Image burnImage, smokeImage;

	// Global scale variable for bounding boxes
	public static float BBscale = 0.8f;

	Font font = Font.font("TimesRoman", FontPosture.ITALIC, 60.0);
	
	// Variables to track scrolling view window
	public static double view_x=0, view_y=0;
	final static double VIEW_WIDTH = 800;
	final static double VIEW_HEIGHT = 600;
	final static double WORLD_WIDTH = 1600;
	final static double WORLD_HEIGHT = 2000;
	final static double SCROLL_X = 100;
	final static double SCROLL_Y = 100;
	// Keep track of which coords the graphics context is
	// currently using
	static boolean world_coords = false;

	
	/**
	 * Set up initial data structures/values
	 */
	void initialize()
	{
		rockImage = new Image("rock.gif");
		lakeImage = new Image("lake.gif");
		treeImage = new Image("tree.gif");
		fireImage = new Image("fire.gif");
		burnImage = new Image("burn.gif");
		smokeImage = new Image("smoke.gif");
		setLevel1();
	}
	
	void setHandlers(Scene scene)
	{
		scene.setOnMousePressed(
			e -> {
				// Adjust x and y coords to allow for view window
				tree.headTo(view_x + e.getX(), view_y + e.getY());
			}
		);		
	}
	

	/**
	 *  Update variables for one time step
	 */
	public void update()
	{
		fire.updateSprite();
		tree.updateSprite();
		checkScrolling(tree.getBoundingBox());
	}
	
	public void checkScrolling(BoundingBox r)
	{
		// Adjust view window so that the rectangle
		// r (in world coords) is inside the
		// scrolling limits (if possible)
		//
		// Check horizontal scrolling
		if (r.getMinX() < view_x + SCROLL_X)
		{
			view_x = r.getMinX() - SCROLL_X;
			if (view_x < 0)
				view_x = 0;
		}
		else if (r.getMaxX() > view_x + VIEW_WIDTH - SCROLL_X)
		{
			view_x = r.getMaxX() - VIEW_WIDTH + SCROLL_X;
			if (view_x +VIEW_WIDTH > WORLD_WIDTH)
				view_x = WORLD_WIDTH - VIEW_WIDTH;
		}
		// Check vertical scrolling
		if (r.getMinY() < view_y + SCROLL_Y)
		{
			view_y = r.getMinY() - SCROLL_Y;
			if (view_y < 0)
				view_y = 0;
		}
		else if (r.getMaxY() > view_y + VIEW_HEIGHT - SCROLL_Y)
		{
			view_y = r.getMaxY() - VIEW_HEIGHT + SCROLL_Y;
			if (view_y +VIEW_HEIGHT > WORLD_HEIGHT)
				view_y = WORLD_HEIGHT - VIEW_HEIGHT;
		}
	}
	
	static void useWorldCoords(GraphicsContext gc)
	{
		if (!world_coords)
		{
			gc.translate(-view_x, -view_y);
			world_coords = true;
		}
	}

	static void useScreenCoords(GraphicsContext gc)
	{
		if (world_coords)
		{
			gc.setTransform(new Affine());
			world_coords = false;
		}
	}

	/**
	 *  Draw the game world
	 */
	void render(GraphicsContext gc) {
		useScreenCoords(gc);
		gc.setFill(Color.WHITE);
		gc.fillRect(0.0, 0.0, VIEW_WIDTH, VIEW_HEIGHT);
		// Reset coords to reflect view window
		useWorldCoords(gc);
		// First layer the dots
		gc.setStroke(Color.BLACK);
		gc.setLineWidth(2.0);
		for (int i = 0; i < dots.length; i++)
			gc.strokeOval(dots[i].getX(), dots[i].getY(), 80, 60);
		//System.out.println("Begin scenery");
		for (Sprite s: scenery)
		{
			s.render(gc, boxDebug);
		}
		fire.render(gc, boxDebug);
		tree.render(gc, boxDebug);
		// Now, reset coords to draw things that
		// don't scroll
		useScreenCoords(gc);

		gc.setFill(Color.BLACK);
		gc.setFont(font);
		gc.fillText("Status Line",200,50);
		if (wonGame())
		{
			gc.fillText("You win!",200,200);
		}

		if (lostGame())
		{
			gc.fillText("You lose.",200,200);
		}
		
	}
	
	public void setLevel1()
	{
		scenery = new Sprite[]{
			new RockSprite(300,200,rockImage),
			new RockSprite(60,400,rockImage),
			new LakeSprite(440,300,lakeImage),
			new RockSprite(240,80,rockImage),
			new RockSprite(1400,1200,rockImage),
			new RockSprite(800,1300,rockImage),
			new RockSprite(200,1740,rockImage),
			new WallSprite(0,0,10,WORLD_HEIGHT-1),
			new WallSprite(0,0,WORLD_WIDTH-1,10),
			new WallSprite(0,WORLD_HEIGHT-10,WORLD_WIDTH-1,10),
			new WallSprite(WORLD_WIDTH-10,0,10,WORLD_HEIGHT-1)
		};

		tree = new TreeSprite(10,10,scenery,treeImage,burnImage);
		fire = new FireSprite(640,200,scenery,tree,fireImage,smokeImage);
		// Create a bunch of random dots to make it easier
		// to follow movement
		dots = new Point2D[40];
		for (int i = 0; i < dots.length; i++)
		{
			double dot_x = Math.random()*WORLD_WIDTH;
			double dot_y = Math.random()*WORLD_HEIGHT;
			dots[i] = new Point2D(dot_x, dot_y);
		}
	}
	
	private boolean lostGame()
	{
		return !tree.isAlive();
	}

	private boolean wonGame()
	{
		return tree.isAlive() && !fire.isAlive();
	}


	/*
	 * Begin boiler-plate code...
	 * [Animation and events with initialization]
	 */
	public static void main(String[] args) {
		launch(args);
	}
	
	@Override
	public void start(Stage theStage) {
		theStage.setTitle(appName);

		Group root = new Group();
		Scene theScene = new Scene(root);
		theStage.setScene(theScene);

		Canvas canvas = new Canvas(VIEW_WIDTH, VIEW_HEIGHT);
		root.getChildren().add(canvas);

		GraphicsContext gc = canvas.getGraphicsContext2D();

		// Initial setup
		initialize();
		setHandlers(theScene);
		
		// Setup and start animation loop (Timeline)
		KeyFrame kf = new KeyFrame(Duration.millis(1000 / FPS),
				e -> {
					// update position
					update();
					// draw frame
					render(gc);
				}
			);
		Timeline mainLoop = new Timeline(kf);
		mainLoop.setCycleCount(Animation.INDEFINITE);
		mainLoop.play();

		theStage.show();
	}
	/*
	 * ... End boiler-plate code
	 */
}
